@@ -1,9 +1,8 @@
 import asyncio
 import logging
-from functools import partial
 from pathlib import Path
 from time import time
-from typing import Coroutine, Optional, Type, TypeVar
+from typing import Optional, Type, TypeVar
 
 import pyppeteer
 from pyppeteer.page import Page
@@ -14,6 +13,7 @@ from scrapy.http import Request, Response
 from scrapy.http.headers import Headers
 from scrapy.responsetypes import responsetypes
 from scrapy.statscollectors import StatsCollector
+from scrapy.utils.defer import deferred_from_coro
 from scrapy.utils.reactor import verify_installed_reactor
 from twisted.internet.defer import Deferred, inlineCallbacks
 
@@ -25,14 +25,12 @@ _patch_pyppeteer_connection()
 del _patch_pyppeteer_connection
 
 
+__all__ = ["ScrapyPyppeteerDownloadHandler"]
+
+
 logger = logging.getLogger("scrapy-pyppeteer")
 
 
-def _force_deferred(coro: Coroutine) -> Deferred:
-    future = asyncio.ensure_future(coro)
-    return Deferred.fromFuture(future)
-
-
 async def _request_handler(
     request: pyppeteer.network_manager.Request, scrapy_request: Request, stats: StatsCollector
 ) -> None:
@@ -41,15 +39,18 @@ async def _request_handler(
         overrides = {
             "method": scrapy_request.method,
             "headers": {
-                key.decode("utf-8"): value[0].decode("utf-8")
+                key.decode("utf-8").lower(): value[0].decode("utf-8")
                 for key, value in scrapy_request.headers.items()
             },
         }
         if scrapy_request.body:
             overrides["postData"] = scrapy_request.body.decode(scrapy_request.encoding)
-        await request.continue_(overrides)
     else:
-        await request.continue_()
+        overrides = {"headers": request.headers.copy()}
+        if scrapy_request.headers.get("user-agent"):
+            user_agent = scrapy_request.headers["user-agent"].decode("utf-8")
+            overrides["headers"]["user-agent"] = user_agent
+    await request.continue_(overrides)
     # increment stats
     stats.inc_value("pyppeteer/request_method_count/{}".format(request.method))
     stats.inc_value("pyppeteer/request_count")
@@ -69,8 +70,11 @@ class ScrapyPyppeteerDownloadHandler(HTTPDownloadHandler):
     def __init__(self, crawler: Crawler) -> None:
         super().__init__(settings=crawler.settings, crawler=crawler)
         verify_installed_reactor("twisted.internet.asyncioreactor.AsyncioSelectorReactor")
-        crawler.signals.connect(self._launch_browser_signal_handler, signals.engine_started)
+        crawler.signals.connect(self._engine_started_handler, signals.engine_started)
         self.stats = crawler.stats
+        self.browser: Optional[pyppeteer.browser.Browser] = None
+
+        # read settings
         self.navigation_timeout: Optional[int] = None
         self.page_coroutine_timeout: Optional[int] = None
         if crawler.settings.get("PYPPETEER_NAVIGATION_TIMEOUT"):
@@ -79,7 +83,6 @@ class ScrapyPyppeteerDownloadHandler(HTTPDownloadHandler):
             self.page_coroutine_timeout = crawler.settings.getint(
                 "PYPPETEER_PAGE_COROUTINE_TIMEOUT"
             )
-        self.browser: Optional[pyppeteer.browser.Browser] = None
         self.launch_options: dict = crawler.settings.getdict("PYPPETEER_LAUNCH_OPTIONS") or {}
         if (
             "executablePath" not in self.launch_options
@@ -92,38 +95,49 @@ class ScrapyPyppeteerDownloadHandler(HTTPDownloadHandler):
     def from_crawler(cls: Type[PyppeteerHandler], crawler: Crawler) -> PyppeteerHandler:
         return cls(crawler)
 
-    def _launch_browser_signal_handler(self) -> Deferred:
-        return _force_deferred(self._launch_browser())
+    def _engine_started_handler(self) -> Deferred:
+        return deferred_from_coro(self._launch_browser())
 
     async def _launch_browser(self) -> None:
         self.browser = await pyppeteer.launch(options=self.launch_options)
 
     def download_request(self, request: Request, spider: Spider) -> Deferred:
         if request.meta.get("pyppeteer"):
-            return _force_deferred(self._download_request(request, spider))
+            return deferred_from_coro(self._download_request(request, spider))
         return super().download_request(request, spider)
 
     async def _download_request(self, request: Request, spider: Spider) -> Response:
+        page = await self._create_page_for_request(request)
         try:
-            page = await self.browser.newPage()  # type: ignore
-            result = await self._download_request_page(request, spider, page)
+            result = await self._download_request_with_page(request, spider, page)
         except Exception:
             if not page.isClosed():
                 await page.close()
+                self.stats.inc_value("pyppeteer/page_count/closed")
             raise
         else:
             return result
 
-    async def _download_request_page(
-        self, request: Request, spider: Spider, page: Page
-    ) -> Response:
+    async def _create_page_for_request(self, request: Request) -> Page:
+        page = await self.browser.newPage()  # type: ignore
         self.stats.inc_value("pyppeteer/page_count")
         if self.navigation_timeout is not None:
             page.setDefaultNavigationTimeout(self.navigation_timeout)
         await page.setRequestInterception(True)
-        page.on("request", partial(_request_handler, scrapy_request=request, stats=self.stats))
-        page.on("response", partial(_response_handler, stats=self.stats))
+        page.on(
+            "request",
+            lambda req: asyncio.create_task(
+                _request_handler(req, scrapy_request=request, stats=self.stats)
+            ),
+        )
+        page.on(
+            "response", lambda resp: asyncio.create_task(_response_handler(resp, stats=self.stats))
+        )
+        return page
 
+    async def _download_request_with_page(
+        self, request: Request, spider: Spider, page: Page
+    ) -> Response:
         start_time = time()
         response = await page.goto(request.url)
 
@@ -153,8 +167,9 @@ class ScrapyPyppeteerDownloadHandler(HTTPDownloadHandler):
                 self.stats.inc_value("pyppeteer/page_count/injected_callback")
                 break
         else:
-            await page.close()
-            self.stats.inc_value("pyppeteer/page_count/closed")
+            if not page.isClosed():
+                await page.close()
+                self.stats.inc_value("pyppeteer/page_count/closed")
 
         headers = Headers(response.headers)
         headers.pop("Content-Encoding", None)
@@ -172,4 +187,4 @@ class ScrapyPyppeteerDownloadHandler(HTTPDownloadHandler):
     def close(self) -> Deferred:
         yield super().close()
         if self.browser:
-            yield _force_deferred(self.browser.close())
+            yield deferred_from_coro(self.browser.close())
